/**
 *  Copyright (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  Licensed under the GPL, Version 3.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/gpl.html
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 * --------------------------------------------------------------------------
 *  版权 (c) 1997-2013, tinygroup.org (luo_guo@live.cn).
 *
 *  本开源软件遵循 GPL 3.0 协议;
 *  如果您不遵循此协议，则不被允许使用此文件。
 *  你可以从下面的地址获取完整的协议文本
 *
 *       http://www.gnu.org/licenses/gpl.html
 */
package org.tinygroup.tree.impl;

import java.util.ArrayList;
import java.util.List;

import org.tinygroup.commons.exceptions.NotExistException;
import org.tinygroup.tree.Tree;

public class TreeImpl<T, E> implements Tree<T, E> {
    private T path = null;
    private E data = null;
    private Tree<T, E> parent = null;
    private List<Tree<T, E>> subTrees = new ArrayList<Tree<T, E>>();

    /**
     * 构造树，根据path
     *
     * @param path 路径
     */
    public TreeImpl(T path) {
        this(path, null);
    }

    /**
     * 构造树，根据path和数据
     *
     * @param path 路径
     * @param data 数据
     */
    public TreeImpl(T path, E data) {
        this.path = path;
        this.data = data;
    }

    /**
     * 获取父结点
     *
     * @return ITree<T,E>
     */

    public Tree<T, E> getParent() {
        return parent;
    }

    /**
     * 获取所有子结点
     *
     * @return List<ITree<T,E>>
     */

    public List<Tree<T, E>> getSubTrees() {
        return subTrees;
    }

    /**
     * 根据路径获取子结点
     *
     * @param path 获取的子路径
     * @return ITree<T,E> 返回命中的节点
     */

    public Tree<T, E> getSubTree(T path) {
        Tree<T, E> t = getSubTreeLocal(path);
        if (t != null) {
            return t;
        }
        throw new NotExistException();
    }

    private Tree<T, E> getSubTreeLocal(T path) {
        for (Tree<T, E> tree : subTrees) {
            if (tree.getPath().equals(path)) {
                return tree;
            }
        }
        return null;
    }

    /**
     * 从当前节点开始，在其子孙中查找指定路径名的子树 返回最先找到的树
     *
     * @param path 路径
     * @return ITree<T,E> 匹配的树节点
     */

    public Tree<T, E> getTree(T path) {
        Tree<T, E> tree = getTreeLocal(path);
        if (tree != null) {
            return tree;
        }
        throw new NotExistException();
    }

    private Tree<T, E> getTreeLocal(T path) {
        if (this.path.equals(path)) {
            return this;
        } else {
            for (Tree<T, E> tree : subTrees) {
                TreeImpl<T, E> treeImpl = (TreeImpl<T, E>) tree;
                TreeImpl<T, E> t = (TreeImpl<T, E>) treeImpl.getTreeLocal(path);
                if (t != null) {
                    return t;
                }
            }
        }
        return null;
    }

    /**
     * 在当前树的指定子树前增加子树 返回所增加的子树
     *
     * @param path 路径
     * @param data 数据
     * @return ITree<T,E> 返回刚才添加的节点
     */

    public Tree<T, E> addTreeOnTail(
            T path, E data) {
        return addTreeOnTail(new TreeImpl<T, E>(path, data));
    }

    public Tree<T, E> addTreeOnTail(Tree<T, E> tree) {
        subTrees.add(tree);
        tree.setParent(this);
        return tree;
    }

    public Tree<T, E> addTreeOnTail(T path) {
        return addTreeOnTail(path, null);
    }

    public Tree<T, E> addTreeOnHead(T path) {
        return addTreeOnHead(path, null);
    }

    public Tree<T, E> addTreeOnHead(T path, E data) {
        return addTreeOnHead(new TreeImpl<T, E>(path, data));
    }

    public Tree<T, E> addTreeOnHead(Tree<T, E> tree) {
        subTrees.add(0, tree);
        tree.setParent(this);
        return tree;
    }

    public Tree<T, E> addTreeBefore(T path, T newPath) {
        return addTreeBefore(path, newPath, null);
    }

    public Tree<T, E> addTreeBefore(T path, Tree<T, E> tree) {
        Tree<T, E> position = this.getSubTree(path);
        if (position != null && position.getParent() != null) {
            for (int i = 0; i < position.getParent().getSubTrees().size(); i++) {
                if (position.getParent().getSubTrees().get(i).getPath()
                        .equals(path)) {
                    tree.setParent(position.getParent());
                    position.getParent().getSubTrees().add(i, tree);
                    break;
                }
            }
        }
        return tree;
    }

    public Tree<T, E> addTreeBefore(T path, T newPath, E data) {
        return addTreeBefore(path, new TreeImpl<T, E>(newPath, data));
    }

    public Tree<T, E> addTreeAfter(T path, T newPath) {
        return addTreeAfter(path, newPath, null);
    }

    public Tree<T, E> addTreeAfter(T path, Tree<T, E> tree) {
        Tree<T, E> position = this.getSubTree(path);
        if (position != null && position.getParent() != null) {
            for (int i = 0; i < position.getParent().getSubTrees().size(); i++) {
                if (position.getParent().getSubTrees().get(i).getPath()
                        .equals(path)) {
                    position.getParent().getSubTrees().add(i + 1, tree);
                    tree.setParent(position.getParent());
                    break;
                }
            }
        }
        return tree;
    }

    public Tree<T, E> addTreeAfter(T path, T newPath, E data) {
        return addTreeAfter(path, new TreeImpl<T, E>(newPath, data));
    }

    public E getData() {
        return data;
    }

    public boolean isLeaf() {
        return subTrees.size() == 0;
    }

    public int getDegree() {
        return subTrees.size();
    }

    public int getHeight() {
        Tree<T, E> tree = this;
        int height = 0;
        while (tree.getParent() != null) {
            height++;
            tree = tree.getParent();
        }
        return height;
    }

    public Tree<T, E> getTreeByPath(T[] path) {
        if (!this.getPath().equals(path[0])) {
            return null;
        } else {
            return getTreeByPath(path, 1);
        }
    }

    private Tree<T, E> getTreeByPath(T[] path, int index) {
        TreeImpl<T, E> tree = (TreeImpl<T, E>) this.getSubTree(path[index]);
        if (path.length - 1 == index) {
            return tree;
        } else {
            return tree.getTreeByPath(path, index + 1);
        }
    }

    public E getDataByPath(T[] path) {
        Tree<T, E> tree = getTreeByPath(path);
        if (tree != null) {
            return tree.getData();
        }
        throw new NotExistException();
    }

    public T getPath() {
        return path;
    }

    public List<T> getFullPath() {
        List<T> paths = new ArrayList<T>();
        Tree<T, E> tree = this;
        while (tree != null) {
            paths.add(0, tree.getPath());
            tree = tree.getParent();
        }
        return paths;
    }

    public void clear() {
        for (Tree<T, E> tree : this.subTrees) {
            tree.clear();
        }
        subTrees.clear();

    }

    public void setParent(Tree<T, E> parent) {
        this.parent = parent;

    }

    public void setData(E data) {
        this.data = data;

    }

    public void removeTree(Tree<T, E> tree) {
        tree.getParent().getSubTrees().remove(tree);

    }

    public void removeTree(T[] path) {
        Tree<T, E> tree = getTreeByPath(path);
        if (tree != null) {
            removeTree(tree);
        }
    }

    public void removeTree(T path) {
        Tree<T, E> tree = getSubTree(path);
        if (tree != null) {
            removeTree(tree);
        }
    }

    public List<T> getSubPaths() {
        List<T> result = new ArrayList<T>();
        for (int i = 0; i < subTrees.size(); i++) {
            Tree<T, E> t = subTrees.get(i);
            result.add(t.getPath());
        }
        return result;
    }

}
